#include <iostream>
#include <list>
#include <gtkmm.h>
using std::cout;
using std::cerr;
using std::endl;
using std::string;
#include "ardour.h"

#define dbg(A, ...) dbgprintf("ayyiArdour", __func__, (char*)A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("ayyiArdour", __func__, (char*)A, ##__VA_ARGS__)
#define err(A, ...) errprintf("ayyiArdour", __func__, (char*)A, ##__VA_ARGS__)
#define PF printf("ayyiArdour::%s()...\n", __func__);
#define SIG_IN printf("%s--->%s %s\n", bold, white, __func__)
#define SIG_IN1 printf("%s--->%s ", bold, white)
#define SIG_OUT printf("%s(): %s--->%s\n", __func__, yellow, white)

extern int debug;

/*
void
AyyiArdour::partlist_sync()
{
	//make sure all local data structs are in sync with the ardour session.
	//-update widget stuff elsewhere.

	//note: currently we dont delete any parts here. Its possible to have partlist items not in shm after this.

	int count = 0;

	//get parts from region POD structs:
	GList* parts = ardour_get_part_list();

	for(; parts; parts = parts->next){
		struct _gpart* gpart = NULL;
		int slot = GPOINTER_TO_INT(parts->data);

		if(!ardour_gpart_exists(slot)) gpart = ardour_gpart_new_from_pod(slot);
		else{
			gpart = ardour_gpart_get_from_region_index(slot);
			ardour_part_sync(gpart);
		}

		if(gpart) count++;
	}
	g_list_free(parts);

	dbg("part count: %u.", count);
}
*/

std::list<region_shared*>
Ayyi_ardour_client::get_parts_list()
{
	/*
		gets ardour regions from the shared_regions struct in shm.

		parts that are 'not being used' (are not in any playlists), are ignored.
	*/

	std::list<region_shared*> part_list;
	struct block* block = region_get_block(0);
	void** region_index = block->slot;
	int i = 0;
	for(i=0;i<=block->last;i++){
		region_shared* region = get_region(i);
		if(!region) continue;
		if(!is_song_shm_local(region)){ err("failed to get shared region struct. region_index[%u]=%p\n", i, region_index[i]); continue; }
		//printf("%s(): region_index[%u]=%p id=%Lu pos=%u.\n", __func__, i, shared, shared->id, shared->position);
		if(region->playlist < 0){ /*printf("ardour_get_part_list(): non playlist region. ignoring...\n");*/ continue; }
		part_list.push_back(region);
	}

	//dbg("num parts found: %i", part_list.size());
	//if(debug) ardour_print_part_list();
	return part_list;
}


region_shared*
Ayyi_ardour_client::get_region(int slot_num)
{
	//returns the shm data for a region with the given slot number.

	if(!region_index_ok(slot_num)){ err("bad region index (%i)\n", slot_num); return NULL; }
	block* block = region_get_block(0);

	if(block->last < 0) dbg("no regions. last=%i", block->last);

	//printf("%s(): block=%p 0=%p\n", __func__, shm_index->regions.block, shm_index->regions.block[0]);

	void** region_index = block->slot;
	if(!is_song_shm_local(region_index)){ err("bad song data. block_index=%p\n", region_index); return NULL; }
	//printf("%s(): slot=%i %p last=%i shm_index=%p local?=%1i\n", __func__, pod_index, region_index, shm_index->regions.last, shm_index, is_song_shm_local(region_index));
	struct _audio_region_shared* region = (struct _audio_region_shared*)xlate_shm_address(region_index[slot_num]);
	if(!region) return NULL;
	if(!is_song_shm_local(region)){ err("failed to get shared region struct. region_index[%u]=%p\n", slot_num, region_index[slot_num]); return NULL; }
	return region;
}


bool
Ayyi_ardour_client::region_index_ok(int pod_index)
{
	//check that the given region slot number is valid.
	//-we use this to ensure pointer dereferencing is ok, so dont return TRUE for any magic index numbers.

	if(pod_index >= AYYI_BLOCK_SIZE) return FALSE;

	int s1, s2;
	struct block* block = region_msg_idx_get_block((unsigned)pod_index, &s1, &s2);
	return (pod_index > -1 && pod_index <= block->last);
}


struct block*
Ayyi_ardour_client::region_msg_idx_get_block(unsigned idx, int* s1, int* s2)
{
    //convert a serialised index into two separate array indexes for the Container and Block.
	//ASSERT_POINTER_NULL(s2, "s2");

    *s1 = idx & 0xffff0000; 
    *s2 = idx & 0x0000ffff;
    return (struct block*)xlate_shm_address(song_shm.data->audio_regions.block[*s1]);
}


