#include <iostream>
#include <list>
#include <gtkmm.h>
using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::list;
#include "utils.h"
#include "wave.h"
#include "ayyi.h"
#include "app.h"

using namespace Ayi;

#define ARDOUR_SOUND_DIR1 "interchange/"
#define ARDOUR_SOUND_DIR2 "audiofiles"

#define dbg(LEVEL, A, ...) dbgprintf2("ayyi", __func__, LEVEL, (char*)A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("ayyi", __func__, (char*)A, ##__VA_ARGS__)
#define err(A, ...) errprintf("ayyi", __func__, (char*)A, ##__VA_ARGS__)
#define PF printf("ayyi::%s()...\n", __func__);

static void dbus_shm_notify   (DBusGProxy*, DBusGProxyCall *call, gpointer _seg);
void        region_deleted_cb (Action*);
void        region_new_cb     (Action*);

Ayyi_client* Ayyi_client::_instance = 0;


template<class T>
Remote_object<T>::Remote_object(T* pod)
	: shared(pod)
{
}


template<class T>
container*
Remote_object<T>::get_container()
{
	return Ayyi_client::instance()->song_shm.data->filesources; //no!
}


template<>
Ayi::container*
Remote_object<filesource_shared>::get_container()
{
	return &Ayyi_client::instance()->song_shm.data->filesources;
}


template<>
container*
Remote_object<route_shared>::get_container()
{
	return &Ayyi_client::instance()->song_shm.data->routes;
}


template<>
container*
Remote_object<_playlist_shared>::get_container()
{
	return &Ayyi_client::instance()->song_shm.data->playlists;
}


template<class T>
Ayi::block*
Remote_object<T>::get_block(int block_num)
{
	return (Ayi::block*)Ayyi_client::instance()->xlate_shm_address(get_container()->block[block_num]);
}


/*
block*
Remote_object<region_shared>::get_block(int block_num)
{
	return (block*)ayyi->xlate_shm_address(ayyi->song_shm.data->regions.block[block_num]);
}


block*
Remote_object<filesource_shared>::get_block(int block_num)
{
	cout << " **********" << endl;
	return (block*)ayyi->xlate_shm_address(ayyi->song_shm.data->filesources.block[block_num]);
}
*/


/*
template<class T>
block*
Remote_object<T>::get_block(container* container, int block_num)
{
	return (block*)ayyi->xlate_shm_address(container->block[block_num]);
}
*/


Ayyi_client::Ayyi_client()
	: trans_id(1), online(false)
{
	_instance = this;

	//currently, the client can only deal with one segment type.
	shm_segs.push_back(&this->song_shm);
}


Ayyi_client::~Ayyi_client()
{
	delete messenger;
}


Ayyi_client*
Ayyi_client::instance()
{
	return _instance;
}


bool
Ayyi_client::connect()
{
	try { messenger = new Dbus; } catch (...){ return false; }

	return true;
}


bool
Ayyi_client::get_shm()
{
	for(std::list<client_shm<struct _shm_song>* >::iterator seg = shm_segs.begin(); seg!=shm_segs.end(); seg++){

		(*seg)->id = messenger->get_shm(seg_strs[(*seg)->type]);
		if((*seg)->import()){
			on_shm_complete();
		}
		else return false;

	}
	return true;
}


void*
Ayyi_client::xlate_shm_address(void* address)
{
	//pointers in shm have to be translated to the address range of the imported segment.

	if(!address) return NULL;

	void* translated = (void*)((uint32_t)address + (uint32_t)song_shm.data - (uint32_t)song_shm.data->owner_shm);

	//if(!is_song_shm(translated)){ errprintf("sml_translate_address(): address is not in shm!\n"); return NULL; }
	return (void*)translated;
}


bool
Ayyi_client::is_song_shm_local(void* p)
{
	//returns true if the pointer is in the local shm segment.
	//returns false if the pointer is in a foreign shm segment.

	unsigned segment_end = (unsigned)song_shm.data + song_shm.size;
	if(((unsigned)p>(unsigned)song_shm.data) && ((unsigned)p<segment_end)) return true;

	err("address is not in song shm: %p (expected %p - %p)\n", p, song_shm.data, (void*)segment_end);
	if((unsigned)p > (unsigned)song_shm.data) err("shm offset=%u\n", (unsigned)p - (unsigned)song_shm.data);
	//else errprintf("is_song_shm(): address is below start of shm segment by %u\n", (unsigned int)shm.song - (unsigned int)p);
	return false;
}


void
Ayyi_client::set_audiodir()
{
	//strictly speaking, i think Ardour files can be anywhere. The path is in the XML file.
	//But we dont currently share the full path, so we have to assume that they are in the Interchange folder.

	Shm_index* shm_index = song_shm.data;
	dbg(2, "samplerate=%i", shm_index->sample_rate);
	char d[256]; snprintf(d, 255, "%s%s%s", shm_index->path, ARDOUR_SOUND_DIR1, shm_index->snapshot);
	dbg(2, "shm_index=%p d=%s", shm_index, d);
	if(!g_file_test(d, G_FILE_TEST_EXISTS)){
		snprintf(d, 255, "%s%s", shm_index->path, ARDOUR_SOUND_DIR1);
		GList* dir_list = get_dirlist(d);
		if(dir_list){
			strncpy(d, (char*)dir_list->data, 255);
			//free the dir list
			GList* list = dir_list; for(;list;list=list->next) g_free(list->data);
			g_list_free(dir_list);
		}
	}

	audiodir = d;
	audiodir += '/';
	audiodir += ARDOUR_SOUND_DIR2;
	audiodir += '/';
	dbg(3, "using audio dir: %s", audiodir.c_str());

	peaksdir = string(shm_index->path) + "peaks/";
}


string*
Ayyi_client::get_audiodir()
{
	return &audiodir;
}


string*
Ayyi_client::get_peaksdir()
{
	return &peaksdir;
}


filesource_shared*
Ayyi_client::get_filesource(int slot_num)
{
	//if(!ardour_region_pod_index_ok(pod_index)){ errprintf("region_pod() bad region index (%i)\n", pod_index); return NULL; }

	/*
	void** filesource_index = song_shm.data->filesources.slot;
	filesource_shared* filesource = (filesource_shared*)xlate_shm_address(filesource_index[slot_num]);
	if(filesource && !is_song_shm_local(filesource)){ err("failed to get shared region struct. region_index[%u]=%p\n", slot_num, filesource_index[slot_num]); return NULL; }
	*/
	Ayi::block* block = Remote_object<filesource_shared>::get_block(0);

	if(block->last < 0) dbg(0, "no files. last=%i", block->last);

	void** table = block->slot;
	if(!is_song_shm_local(table)){ err("bad song data. filesource_table=%p\n", table); return NULL; }
	filesource_shared* filesource = (filesource_shared*)xlate_shm_address(table[slot_num]);
	if(!filesource) return NULL;
	if(!is_song_shm_local(filesource)){ err("failed to get shared filesource struct. filesource_table[%u]=%p\n", slot_num, table[slot_num]); return NULL; }
	return filesource;
}


std::list<filesource_shared*>
Ayyi_client::files_get_list()
{
	static std::list<filesource_shared*> file_list;
	file_list.clear();

	Ayi::block* block = Remote_object<filesource_shared>::get_block(0);
	void** filesource_index = block->slot;
	//printf("%s(): filesources.slot=%p\n", __func__, filesources.slot);
	if(block->last >= AYYI_BLOCK_SIZE){ err("last=%i\n", block->last); return file_list; }
	int i = 0;
	for(i=0;i<=block->last;i++){
		if(!filesource_index[i]) continue;
		filesource_shared* shared = get_filesource(i);
		if(!is_song_shm_local(shared)){ err("failed to get shared region struct. region_index[%u]=%p\n", i, filesource_index[i]); continue; }
		//printf("%s(): file_index[%u]=%p %14s src_id=%4Lu length=%u.\n", __func__, i, shared, shared->name, shared->id, shared->length);
		//if(!strlen(shared->name)){ /*printf("ardour_get_files_list_pod(): non playlist region. ignoring...\n");*/ continue; }

#ifdef ARDOURD_WORKS_PROPERLY
		char msg[256]; snprintf(msg, 255, "file has zero length! %s", shared->name);
		if(!shared->length) log_append(msg, LOG_WARN);
#endif

		if(shared->id == 0){ err("bad song data found. File has illegal source_id 0."); continue; }

		file_list.push_back(shared);
	}

	dbg(2, "num files found: %i", file_list.size());
	return file_list;
}


Partlist
Ayyi_client::get_parts_list()
{
}


route_shared*
Ayyi_client::get_track(int slot_num)
{
	//if(!ardour_track_pod_index_ok(pod_index)){ P_ERR ("bad track index (%i) last=%i\n", pod_index, shm_index->routes.last); return NULL; }

	block* block = Remote_object<route_shared>::get_block(0);
	void** route_index = block->slot;
	route_shared* route = (route_shared*)xlate_shm_address(route_index[slot_num]);
	if(!route) return NULL; //no route has been allocated for this index.
	if(!is_song_shm_local(route)){ err("failed to get shared route struct. route_index[%u]=%p\n", slot_num, route_index[slot_num]); return NULL; }
	return route;
}


_playlist_shared*
Ayyi_client::get_playlist(int slot_num)
{
	//if(!ardour_track_pod_index_ok(pod_index)){ P_ERR ("bad track index (%i) last=%i\n", pod_index, shm_index->routes.last); return NULL; }

	block* block = Remote_object<_playlist_shared>::get_block(0);
	void** route_index = block->slot;
	_playlist_shared* route = (_playlist_shared*)xlate_shm_address(route_index[slot_num]);
	if(!route) return NULL; //no route has been allocated for this index.
	if(!is_song_shm_local(route)){ err("failed to get shared route struct. route_index[%u]=%p\n", slot_num, route_index[slot_num]); return NULL; }
	return route;
}


region_shared*
Ayyi_client::get_region(int slot_num)
{
}


region_shared*
Ayyi_client::region_next(region_shared* prev)
{
	//not tested!!

	region_shared* region = NULL;
	int slot_num  = prev->shm_idx % AYYI_BLOCK_SIZE;
	int block_num = prev->shm_idx / AYYI_BLOCK_SIZE;

	container* container = &song_shm.data->audio_regions;
	if(container->last < 0) return region; //no regions.

	bool first_block = true;
	for(int b=block_num; b<=container->last; b++){
		struct block* block = (struct block*)xlate_shm_address(container->block[b]);
		void** table = block->slot;

		int i = first_block ? slot_num+1 : 0;
		int size = MIN(AYYI_BLOCK_SIZE, block->last+1);
		for(; i<size; i++){
			//if(i > block->last) break;
			if(!table[i]) continue;
			region = (region_shared*)xlate_shm_address(table[i]);
			if(!is_song_shm_local(region)){ err("bad shm pointer. index[%u]=%p\n", i, table[i]); return region=NULL; }
			return region;
		}

		first_block = false;
	}
	dbg(0, "next region not found.");
	return region;
}


bool
Ayyi_client::filename_get_from_region(std::string* filename, region_shared* region)
{
	uint64_t source = region->source0;

	block* block = Remote_object<filesource_shared>::get_block(0);
    void** file_table = block->slot;
	for(int f=0; f<=block->last; f++){
    	if(!file_table[f]) continue;
		filesource_shared* s = get_filesource(f);
		dbg(2, "%i: %s", f, s->name);
		if(s->id == source){
			*filename = s->name;
			return true;
		}
	}
	return false;
}


int
Ayyi_client::region_get_track_num(int region_idx)
{
	region_shared* region = get_region(region_idx);
	//if(!strlen(region->playlist_name)){ warn("unable to determine track. region has no playlist! (%i)\n", region_idx); return 0; }

	_playlist_shared* playlist = get_playlist(region->playlist);
	int t = tracknum_get_from_playlist(playlist);

	return t;
}


int
Ayyi_client::tracknum_get_from_playlist(_playlist_shared* playlist)
{
	/*
	Returns the server track index for the given playlist.

	The track name appears to be the playlist name without the ".1" at the end.
	Ardour can have multiple playlists per track, but we only support one.
	*/

	if(!strlen(playlist->name)){ err("bad arg: empty playlist_name.\n", ""); return 0;}

	gchar* dot = g_strrstr(playlist->name, ".");
	if(!dot){
		dbg(1, "playlist_name has no dot. Assuming this is legal...", "");
		dot = &playlist->name[strlen(playlist->name)];
	}
	int str_len = (unsigned)dot - (unsigned)playlist->name;
	char track_name[AYYI_NAME_MAX];
	strncpy(track_name, playlist->name, str_len);
	track_name[str_len] = '\0';

	int t = 0;
	block* block = Remote_object<route_shared>::get_block(0);
    void** route_table = block->slot;
	for(t=0;t<=block->last;t++){           //FIXME use iterator
        if(!route_table[t]) continue;
		route_shared* route = get_track(t);
		dbg(3, "  testing trk=%s.", route->name);
		if(!strcmp(track_name, route->name)){
			dbg(2, "tnum=%i.", t);
			return t;
		}
	}
	warn("track not found (%s).\n", track_name);
	return t;
}


int
Ayyi_client::get_track_count()
{
    int track_count = 0;
	block* block = Remote_object<route_shared>::get_block(0);
    void** route_table = block->slot;
    for(int i=0;i<=block->last;i++){
        if(!route_table[i]) continue;
        track_count++;
    }
    return track_count;
}


void
Ayyi_client::replace_region_with_multiple(Region* region, std::list<song_object*> new_regions)
{
	// 1- remove old region
	// 2- add multiple new regions.
	PF;

	Action* action   = new Action("delete region");
	action->obj_idx  = region->get_idx();
	action->callback = region_deleted_cb;
	action->op       = AYYI_DEL;
	action->obj_type = AYYI_OBJECT_AUDIO_PART;

	//add more actions to be sent after the first is complete:

	Action* action2 = 0;
	for(std::list<song_object*>::iterator i = new_regions.begin(); i != new_regions.end(); i++){

		action2 = new Action("add region");
		action2->callback = region_new_cb;
		action2->op       = AYYI_NEW;
		action2->obj_type = AYYI_OBJECT_AUDIO_PART;
		action2->obj      = *i;
		action2->pending.push_back(action);
	}

	action->execute();
}


void
Ayyi_client::part_move(Region* region, const gpos* pos)
{
	//this is not tested!

	Action* action   = new Action("part move");
	action->op       = AYYI_SET;
	action->obj_type = AYYI_OBJECT_PART;
	action->obj_idx  = region->get_idx();
	//action->obj      = region;
	//action->i_val    = pos2samples(pos); //FIXME fn missing

	action->execute();
}


uint64_t
Ayyi_client::samples2mu(uint32_t samples)
{
	//returns the length in high resolution tempo-related units, of the given number of audio samples.

	//normally, the ratio is of the order of 1924

	int sample_rate = song_shm.data->sample_rate;
	if(!sample_rate){ err("samplerate zero!\n", __func__); sample_rate = 44100; }

	float secs = (float)samples / (float)sample_rate; //length in seconds.

	float bpm = song_shm.data->bpm;
	float length_of_1_beat_in_secs = 60.0/bpm;

	float beats = secs / length_of_1_beat_in_secs;

	uint64_t mu = beats2mu(beats);

	//FIXME this check below indicates there are some rounding errors:
	//dbg(0, "samples=%Li mu=%Li check=%Li.", samples, mu, mu2samples(mu));
	dbg(0, "samples=%lu mu=%Li secs=%.3f.", samples, mu, secs);
	return mu;
}


void
Ayyi_client::samples2pos(struct _song_pos* pos, uint32_t samples)
{
  int64_t mu = samples2mu(samples);
  mu2pos(mu, (struct ::_song_pos*)pos);
}


bool
Ayyi_client::gpos_is_valid(gpos* pos)
{
  //ASSERT_POINTER_FALSE(pos, "pos");

  if(pos->sub  > 4                ){ err("sub out of range (%i)\n",   pos->sub);  pos->sub  = 3; }
  if(pos->tick > TICKS_PER_SUBBEAT){ err("ticks out of range (%i)\n", pos->tick); pos->tick = TICKS_PER_SUBBEAT-1; }
  return TRUE;
}


void
Ayyi_client::on_shm_complete()
{
	set_audiodir();
	online = true;
}


struct block*
Ayyi_client::region_get_block(int block_num)
{
    return (struct block*)xlate_shm_address(song_shm.data->audio_regions.block[block_num]);
}


Ayyi_client::Region::Region(int shm_idx)
	: _slot_num(shm_idx)
{
	shm_block = Ayyi_client::instance()->region_get_block(0);
	_region   = Ayyi_client::instance()->get_region(shm_idx);
}


uint32_t
Ayyi_client::Region::get_pos()
{
	return _region->position;
}


const char*
Ayyi_client::Region::get_name()
{
	return _region->name;
}


uint32_t
Ayyi_client::Region::get_track_num()
{
	return Ayyi_client::instance()->region_get_track_num(_slot_num); //FIXME ugly
}


static void
dbus_shm_notify(DBusGProxy *proxy, DBusGProxyCall *call, gpointer _seg)
{
	//our request for shm address has returned.
	PF;

	client_shm<struct _shm_song>* seg;
	guint id;
	if(call){
		GError *error = NULL;
		if (!dbus_g_proxy_end_call (proxy, call, &error, G_TYPE_INVALID, G_TYPE_UINT, &id, G_TYPE_INVALID)){
			err("failed to get shm address: %s\n", error->message);
			g_error_free(error);
			return;
		}
	}

	seg->import();
	if(1/*!app.got_song_shm && app_shm_setup_is_complete()*/) Ayyi_client::instance()->on_shm_complete();
}


void
region_deleted_cb(Action* action)
{
	PF;
	if(action) delete action;
}


void
region_new_cb(Action* action)
{
	PF;
	if(action) delete action;
}


