#include <jack/jack.h>
#include <dbus/dbus-glib-bindings.h>
#include "ayyi_types.h"
#include "interface.h"
#include "message_dbus.h"
#include "shm.h"
#include "utils.h"

namespace Ayi {
	typedef struct _container             container;
	typedef struct _filesource_shared     filesource_shared;
	typedef struct _audio_region_shared   region_shared;
	//typedef struct _ayyi_channel          track_shared;
}
using namespace Ayi;
typedef std::list<filesource_shared*> Filelist;
typedef std::list<region_shared*>     Partlist;


struct _gsong_pos
{
  int                beat;
  unsigned short int sub;           //subbeat (4 per beat) - note: not the same as core sub.
  unsigned short int tick;          //960 per sub, as displayed. Not the same as core mu.
};
typedef struct _gsong_pos gpos;


template<class T>
class Remote_object
{
  public:
	Remote_object(T* pod);
	static Ayi::container* get_container();
	static Ayi::block* get_block(int block_num);
	//static block* get_block(container* container, int block_num);

  protected:
	T*            shared;
};


template<class T>
class Remote_object_list
{
  public:
	class iterator : public std::iterator<std::output_iterator_tag,T> {
	  public:
		T& operator*() const { return remote_object; } 
	  private:
		T* remote_object;
	};
};


class Ayyi_client
{
  public:
	                              Ayyi_client();
	                             ~Ayyi_client();
	static Ayyi_client*           instance();
	bool                          connect();
	bool                          get_shm();
	void                          on_shm_complete();
	void*                         xlate_shm_address(void* address);
	bool                          is_song_shm_local(void* p);
	std::string*                  get_audiodir();
	std::string*                  get_peaksdir();

	virtual filesource_shared*    get_filesource(int slot_num);
	virtual Filelist              files_get_list();
	virtual Partlist              get_parts_list();
	Ayi::route_shared*            get_track(int slotnum);
	Ayi::_playlist_shared*        get_playlist(int slot_num);
	virtual region_shared*        get_region(int slot_num);
	region_shared*                region_next(region_shared*);
	virtual bool                  filename_get_from_region(std::string* filename, region_shared*);
	int                           region_get_track_num(int region_idx);
	int                           tracknum_get_from_playlist(_playlist_shared*);
	int                           get_track_count();

	class Region
	{
	  public:
		Region(int shm_idx);
		uint32_t                  get_idx() { return _slot_num; }
		uint32_t                  get_pos();
		uint32_t                  get_track_num();
		const char*               get_name();
	  protected:
		uint32_t                  _slot_num;
		region_shared*            _region;
		Ayi::block*               shm_block;
	};

	//convenience functions for modifying server objects:
	void                          replace_region_with_multiple(Region* region, std::list<song_object*> new_regions);
	void                          part_move(Region* region, const gpos*);

	Messenger*                    messenger;

	std::list<client_shm<Ayi::_shm_song>* > shm_segs;
	client_shm<Ayi::_shm_song> song_shm;             //as song is used 99% of the time, we make a shortcut for it.
	std::string audiodir;
	std::string peaksdir;

	bool                          online;

	unsigned                      trans_id;
	std::list<Action*>            actionlist;

	uint64_t                      samples2mu(uint32_t samples);
	void                          samples2pos(struct _song_pos* pos, uint32_t samples);
	static bool                   gpos_is_valid(gpos*);

	enum{                         TICKS_PER_SUBBEAT = 384,
	                              TICKS_PER_BEAT    = 4*384,
	                              PX_PER_BEAT       = 8,
	    };
  protected:
	static Ayyi_client*           _instance;
	struct block*                 region_get_block(int block_num);
	void                          set_audiodir();
};


