#include <sys/time.h>
#include <iostream>
#include <list>
#include <gtkmm.h>
using std::cout;
using std::cerr;
using std::endl;
using std::string;
#include "ayyi.h"

#include "dbus/dbus_marshallers.h"
#include "dbus/dbus_marshallers.c"

#define ARDOURD_DBUS_SERVICE   "org.ayyi.ardourd.ApplicationService"
#define ARDOURD_DBUS_PATH      "/org/ayyi/ardourd/Ardourd"      //the "node" tag in the xml file, and string in the dbus_g_connection_register_g_object() call.
#define ARDOURD_DBUS_INTERFACE "org.ayyi.ardourd.Application"   //the "interface" name given in the xml file

#define dbg(LEVEL, A, ...) dbgprintf2("messenger", __func__, LEVEL, (char*)A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("messenger", __func__, (char*)A, ##__VA_ARGS__)
#define err(A, ...) errprintf("messenger", __func__, (char*)A, ##__VA_ARGS__)
#define PF printf("messenger::%s()...\n", __func__);
#define P_IN printf("%s--->%s ", green, white)
#define SIG_IN printf("%s--->%s %s\n", bold, white, __func__)
#define SIG_IN1 printf("%s--->%s ", bold, white)
#define SIG_OUT printf("%s(): %s--->%s\n", __func__, yellow, white)

DBusGProxy *proxy;
DBusGProxy *session_proxy = NULL;

extern Ayyi_client* ayyi;

static void dbus_object_new     (Action*, unsigned char type, char* name, gboolean from_source, uint32_t src_idx, guint32 parent_idx, struct _song_pos* stime, guint64 len, guint32 inset);
static void dbus_object_new     (Action*);
static void create_object_notify(DBusGProxy*, DBusGProxyCall*, gpointer data);
static void dbus_object_del     (Action*, int object_type, uint32_t object_idx);
static void delete_object_notify(DBusGProxy*, DBusGProxyCall*, gpointer data);
static void dbus_rx_transport   (DBusGProxy*, int type, gpointer user_data);
static void dbus_rx_property    (DBusGProxy*, int type, gpointer user_data);
static void dbus_rx_progressbar (DBusGProxy*, double val, gpointer user_data);
static void hello_signal_handler(DBusGProxy*, const char *hello_string, gpointer user_data);



song_object::song_object(string obj_name)
	: name(obj_name), from_source(false), src_idx(0), parent_idx(0), len(0)
{
}


Messenger::Messenger()
{
	#define BUS_COUNT 2
	struct _busses {
		int bus;
		char name[64];
	} busses[BUS_COUNT] = {
		{DBUS_BUS_SESSION, "session"},
		{DBUS_BUS_SYSTEM, "system"}
	};

	//the DBusGConnection should never be unreffed, it lives once and is shared amongst the processes.
	bool have_service = false;
	DBusGConnection *connection;
	int i; for(i=0;i<BUS_COUNT;i++){
		GError *error = NULL;
		dbg(2, "trying bus: '%s'...", busses[i].name);
		connection = dbus_g_bus_get((DBusBusType)busses[i].bus, &error);
		if(!connection){
			printf("failed to open connection to %s%s%s bus", bold, busses[i].name, white);
			if(error->code != 0){
				switch(error->code){
					case 23:
						//normal error when bus is not running. Dont report.
						break;
					default:
						printf(": %i: %s\n", error->code, error->message);
						break;
				}
			}
			//printf("\n");
			g_error_free (error);
		}else{
			if((proxy = dbus_g_proxy_new_for_name (connection, ARDOURD_DBUS_SERVICE, ARDOURD_DBUS_PATH, ARDOURD_DBUS_INTERFACE))){
				//it seems we always get a proxy, even if the requested service isnt running...
				//dbg("got proxy ok", "");
				if(service_test()){ have_service = TRUE; break; }
			}
			else dbg(0, "service not running on this bus? (%s)", busses[i].name);
		}
	}
	if(!connection  ){ err("no dbus found.\n", "");           throw std::exception(); }
	if(!have_service){ err("no service found on bus.\n", ""); throw std::exception(); }
	dbg(0, "using '%s' bus.", busses[i].name);
	printf("server messaging service connection\n%s%s\n", go_rhs, ok);

	//also connect to session bus
	DBusGConnection *session_connection;
	if(FALSE){
		GError *error = NULL;
		session_connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
		if(session_connection){
			session_proxy = dbus_g_proxy_new_for_name (session_connection,
							 "org.designfu.TestService",
							 "/org/designfu/TestService/object",
							 "org.designfu.TestService");
			if(!session_proxy){
				err("Failed to get name owner\n", error->message);
			}

		}else{
			err("failed to open connection to %ssession%s bus: %s\n", bold, white, error->message);
			g_error_free (error);
			error = NULL;
		}
	}

	dbus_g_object_register_marshaller (_ad_dbus_marshal_VOID__STRING_INT, G_TYPE_NONE, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INVALID);
	dbus_g_object_register_marshaller (_ad_dbus_marshal_VOID__INT_INT, G_TYPE_NONE, G_TYPE_INT, G_TYPE_INT, G_TYPE_INVALID);

	//tell the proxy which signals we are interested in.

	dbus_g_proxy_add_signal     (proxy, "TransportSignal", G_TYPE_STRING, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (proxy, "TransportSignal", G_CALLBACK (dbus_rx_transport), NULL, NULL);

	dbus_g_proxy_add_signal     (proxy, "PropertySignal", G_TYPE_STRING, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (proxy, "PropertySignal", G_CALLBACK (dbus_rx_property), NULL, NULL);

	dbus_g_proxy_add_signal     (proxy, "ProgressbarSignal", G_TYPE_DOUBLE, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (proxy, "ProgressbarSignal", G_CALLBACK (dbus_rx_progressbar), NULL, NULL);

	dbus_g_proxy_add_signal     (proxy, "HelloSignal", G_TYPE_STRING, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (proxy, "HelloSignal", G_CALLBACK (hello_signal_handler), NULL, NULL);

	dbus_g_proxy_call_no_reply  (proxy, "emitHelloSignal", G_TYPE_INVALID);
}


int
Messenger::get_shm(const char* seg_name)
{
}


bool
Messenger::service_test()
{
	if(!proxy) return false;

	GError *error = NULL;
	int id;
	if(!dbus_g_proxy_call(proxy, "GetShmSingle", &error, G_TYPE_STRING, seg_strs[SEG_TYPE_SONG], G_TYPE_INVALID, G_TYPE_UINT, &id, G_TYPE_INVALID)){
		err("GetShm: %s\n", error->message);
		g_error_free(error);
		return false;
	}
	return true;
}


bool
Messenger::send(Action* action)
{
	return true;
}


Action::Action(std::string description)
{
	id = ayyi->trans_id++;
	snprintf(label, 63, "%s", description.c_str());
	gettimeofday(&time_stamp, NULL);

	ayyi->actionlist.push_back(this);
	printf("%sAction::%s%s(): len=%i\n", yellow, __func__, white, ayyi->actionlist.size());
}


Action::~Action()
{
	ayyi->actionlist.remove(this);

	//check no other messages are waiting for this one.

	if(!ayyi->actionlist.size()) dbg(0, "no other actions to check.");

	for(std::list<Action*>::iterator i=ayyi->actionlist.begin(); i!=ayyi->actionlist.end(); i++){
		Action* action = (*i);
		dbg(0, "action %i: pending actions: %i", action->id, action->pending.size());
		if(action->pending.size()){
			//if the completed action is in the list, remove it.
			for(std::list<Action*>::iterator j=action->pending.begin(); j!=action->pending.end(); j++){
				dbg(0, "pending item... ");
				if(*j != this) continue;

				dbg(0, "erasing...");
				action->pending.erase(j);
				break;
			}
			action->execute();
		}
	}
	dbg(0, "done. actionlist.size=%i", ayyi->actionlist.size());
}


void
Action::execute()
{
	ayyi->messenger->send(this);

	switch(op){
		case AYYI_NEW:
			switch(obj_type){
				case AYYI_OBJECT_PART:
					dbg(0, "AYYI_NEW: PART");
					//dbus_object_new(this, obj_type, (char*)obj->name.c_str(), obj->from_source, obj->src_idx, obj->parent_idx, &obj->stime, obj->len);
					dbus_object_new(this);
					break;
				default:
					err("AYYI_NEW: unexpected object type: %i\n", obj_type);
					break;
			}
			break;
		case AYYI_DEL:
			switch(obj_type){
				case AYYI_OBJECT_PART:
					dbg(0, "DELETE: PART");
					dbus_object_del(this, obj_type, obj_idx);
					break;
				default:
					err("DELETE: unexpected object type: %i\n", obj_type);
					break;
			}
			break;
		default:
			err("unexpected operation: %i\n", op);
			break;
	}
}


void
Action::complete()
{
  if(callback){
	(*(callback))(this);
	//the transaction and its struct are free'd in the above callback.
  }
  else{
	delete (this);
	//printf("  transactionlist tidy finished.\n");
  }
}


static void
dbus_object_new(Action* action, unsigned char type, char* name, gboolean from_source, uint32_t src_idx, guint32 parent_idx, struct _song_pos* stime, guint64 len, guint32 inset)
{
	SIG_OUT;
	guint64 stime64; //TODO send the struct directly.
	if (stime) memcpy(&stime64, stime, 8);

	gpointer user_data = action;
	dbus_g_proxy_begin_call(proxy, "CreateObject", create_object_notify, user_data, NULL, 
							G_TYPE_INT, type, G_TYPE_STRING, name, G_TYPE_BOOLEAN, from_source, G_TYPE_UINT, src_idx, 
							G_TYPE_UINT, parent_idx, G_TYPE_UINT64, stime64, G_TYPE_UINT64, len, G_TYPE_UINT, inset, G_TYPE_INVALID);
}


static void
dbus_object_new(Action* action)
{
	SIG_OUT;
	song_object* obj = action->obj;

	guint64 stime64; //TODO send the struct directly.
	if (&obj->stime) memcpy(&stime64, &obj->stime, 8);

	dbus_g_proxy_begin_call(proxy, "CreateObject", create_object_notify, action, NULL, 
							G_TYPE_INT, obj->obj_type, G_TYPE_STRING, obj->name.c_str(), G_TYPE_BOOLEAN, obj->from_source, G_TYPE_UINT, obj->src_idx, 
							G_TYPE_UINT, obj->parent_idx, G_TYPE_UINT64, stime64, G_TYPE_UINT64, obj->len, G_TYPE_UINT, obj->inset, G_TYPE_INVALID);
}


static void
create_object_notify(DBusGProxy *proxy, DBusGProxyCall *call, gpointer data)
{
	P_IN;
	PF;
	Action* action = (Action*)data;
	if(!action) warn("no action!\n", "");

	guint idx;
	GError *error = NULL;
	if (!dbus_g_proxy_end_call (proxy, call, &error, G_TYPE_UINT, &idx, G_TYPE_INVALID)){
		err("%s\n", error->message);
		g_error_free(error);
		return;
	}
	dbg(0, "request done. idx=%u action=%i", idx, action->id);

	if(action){
      action->idx = idx;
      action->complete();
    }
}


void
dbus_object_del(Action *action, int object_type, uint32_t object_idx)
{
	SIG_OUT;
	gpointer user_data = action;
	dbus_g_proxy_begin_call(proxy, "DeleteObject", delete_object_notify, user_data, NULL, 
							G_TYPE_INT, object_type, G_TYPE_UINT, object_idx, G_TYPE_INVALID);
}


static void
delete_object_notify(DBusGProxy *proxy, DBusGProxyCall *call, gpointer data)
{
	P_IN;
	PF;
	Action* action = (Action*)data;
	if(!action) warn("no action!\n", "");

	//guint idx;
	GError *error = NULL;
	if (!dbus_g_proxy_end_call (proxy, call, &error, G_TYPE_INVALID)){
		err("%s\n", error->message);
		g_error_free(error);
		return;
	}
	dbg(2, "request done. action=%i\n", action->id);

	action->complete();
}


static void
dbus_rx_transport(DBusGProxy *proxy, int type, gpointer user_data)
{
	SIG_IN;
}


static void
dbus_rx_property(DBusGProxy *proxy, int type, gpointer user_data)
{
	SIG_IN;
}


static void
dbus_rx_progressbar(DBusGProxy *proxy, double val, gpointer user_data)
{
	//SIG_IN;
	if(val >= 0.0 && val <= 1.0) {};
}


static void
hello_signal_handler(DBusGProxy *proxy, const char *hello_string, gpointer user_data)
{
	SIG_IN1;
	printf ("Core says: %s\n", hello_string);
}


