#include <iostream>
#include <sndfile.h>
#include <gtkmm.h>
using std::cout;
using std::cerr;
using std::endl;
using std::string;
#include "ardour.h"
#include "wave.h"

#define dbg(A, ...) dbgprintf("wave", __func__, (char*)A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("wave", __func__, (char*)A, ##__VA_ARGS__)
#define err(A, ...) errprintf("wave", __func__, (char*)A, ##__VA_ARGS__)
#define PF printf("wave::%s()...\n", __func__);

extern Ayyi_client* ayyi;

Wave::Wave(const char* filename)
	: sndfile(0)
{
	//currently we open the file imediately, as the peak generator needs to know the length of the file.
	path = filename;
	if(!open()) throw std::exception();
}


Wave::Wave(int region_idx)
	: sndfile(0)
{
	region_shared* region = ayyi->get_region(region_idx);
	if(!region) throw std::exception();

	ayyi->filename_get_from_region(&path, region);
	if(!open()) throw std::exception();

	region_start = region->start;
	region_end   = region_start + region->length;
	if(region_end > frames) throw std::exception();
}


Wave::~Wave()
{
	if(sndfile) sf_close(sndfile);
}


bool
Wave::open()
{
    SF_INFO sfinfo;
    memset(&sfinfo, 0, sizeof(SF_INFO));

	char fullpath[256]; get_full_path(fullpath);

    sndfile = sf_open(fullpath, SFM_READ, &sfinfo);
    if (!sndfile) {
        cerr << "Wave::open(): ERROR: Failed to open input file \"" << fullpath << "\": " << sf_strerror(sndfile) << endl;
        return false;
    }

	frames     = sfinfo.frames;
	samplerate = sfinfo.samplerate;
	channels   = sfinfo.channels;

	return true;
}


void
Wave::get_full_path(char* fullpath)
{
	if(!ayyi->online){ err(""); return; }

	if(path[0]=='/'){
		//absolute path.
		strncpy(fullpath, path.c_str(), 255);
	}else{
		//relative paths are relative to the song default audio directory.
		//if(!ayyi->audiodir.size()){ err(""); return; }
		snprintf(fullpath, 255, "%s/%s", ayyi->audiodir.c_str(), path.c_str());
	}
}


void
Wave::get_full_path(std::string* fullpath)
{
	if(!ayyi->online){ err(""); return; }

	if(path[0]=='/'){
		//absolute path.
		*fullpath = path;
	}else{
		//relative paths are relative to the song default audio directory.
		*fullpath = ayyi->audiodir + path;
	}
}

