
#include <gtkglmm.h>
#include "rulerbar.h"

typedef std::map<const int, std::list<Vamp::RealTime> >::iterator Resultset_iter;


class GLScene : public Gtk::DrawingArea,
                public Gtk::GL::Widget<class GLScene>
{
  public:
	                              GLScene();
	virtual                      ~GLScene();

	void                          load_texture(int texture_name, GdkPixbuf* pixbuf);
	void                          update_lines();
	void                          redraw();
	Glib::RefPtr<Gdk::GL::Window> glwindow;
	double                        get_frames_per_px();

	void                          print(double x, double y, const char* fmt, ...);

	double                        pos2px(gpos*);
	static double                 px2beat(double px);
	bool                          on_client_event(GdkEventClient*);
	bool                          on_button_press_event(GdkEventButton*);
  protected:
	virtual void                  on_realize();
	virtual bool                  on_configure_event(GdkEventConfigure* event);
	virtual bool                  on_expose_event(GdkEventExpose* event);
	void                          start_picking(int cursor_x, int cursor_y);

  protected:
	// font rendering stuff:
	static const Glib::ustring    m_FontString;
	GLuint                        m_FontListBase;
	int                           m_FontHeight;

	Rulerbar                      rulerbar;
};

///////////////////////////////////////////////////////////////////////////////
//
// The application class.
//
///////////////////////////////////////////////////////////////////////////////

class AppWindow : public Gtk::Window
{
  public:
	                 AppWindow();
	virtual         ~AppWindow();
	void             print_map();
	void             on_aubio_init();

	GLScene          glScene;
	Peak*            peak;
	std::map<int, std::list<Vamp::RealTime> > result_cache;
	Resultset_iter   current_lineset;
  protected:
	// signal handlers:
	void             on_button_cancel_clicked();
	void             on_button_ok_clicked();
	void             on_threshold_slider_changed();

protected:
	// member widgets:
	Gtk::Table        table;
	Gtk::VBox         m_VBox;
	Gtk::HBox         hbox;
	Gtk::VScale       threshold_slider;
	Gtk::HScale       zoom_slider;
	Gtk::ComboBoxText onset_type_selector;
	Gtk::ProgressBar  progress_bar;
	Gtk::Button       button_cancel;
	Gtk::Button       button_ok;
};

